/* (C) 2003 Willem Jan Hengeveld  itsme@xs4all.nl */
#include <windows.h>
#include "ItsUtils.h"
#include "dllversion.h"

#include "debug.h"
#include "args.h"
#include <stdio.h>
#include <string.h>
#include <rapi.h>

#include <vector>
using namespace std;

char *msgFlashStartErase= "FLASH: Start erase address %xh\r\n";
char *msgFlashErase= "FLASH: +FlashErase Start=%Xh, Len=%Xh\n";

// to flash the bootloader, we have to unprotect it:
//

// look where the strings 'FLASH: Start erase address' and 'FLASH: +FlashErase
// Start' are referenced, between these there should be 4 consequetive RAM
// addresses, the lowest is the bootloader address
// these 4 addresses contain: 
//       bootloaderstart, bootloadersize, kernelstart, kernelsize.

struct versiondata {
    DWORD dwVersion;
    WCHAR *versionstring;
    DWORD dwAddressList;
    DWORD dwBootloaderAddress;
} versions[]= {
    { 0x80041918, L"3.04.00 ENG", 0x8005FDCC, 0x8C0C60E8   },
    { 0x80055434, L"3.16.52 ENG", 0x80086610, 0x8C0BA204   },
    { 0x80055434, L"3.17.03 ENG", 0x80086648, 0x8C0BA204   },
    { 0x80055414, L"3.19.01 GER", 0x80086628, 0x8C0BA204   },
    { 0x80055be4, L"4.00.10 ENG", 0x80096F48, 0x8C0D62D8   },
    { 0x80055be4, L"4.00.01 ENG", 0x80096184, 0x8C0D62D8   },
    { 0x80055be4, L"4.00.05 ENG", 0x80096F48, 0x8C0D62D8   },
};
#define N_VERSIONS (sizeof(versions)/sizeof(*versions))

bool ITFlashROM(DWORD dwOffset, BYTE *buffer, DWORD dwBufLength);
bool ITWriteProcessMemory(HANDLE hProc, DWORD dwOffset, BYTE *buffer, DWORD dwBufLength, DWORD *pdwNumberOfBytesWritten);
bool ITReadProcessMemory(HANDLE hProc, DWORD dwOffset, BYTE *buffer, DWORD dwBytesWanted, DWORD *pdwNumberOfBytesRead);

bool DisableWriteProtect(DWORD dwFlashInfoOffset);
bool EnableWriteProtect(DWORD dwFlashInfoOffset);
bool FindFlashInfoOffset(DWORD& dwFlashInfoOffset);

bool ReadFromFile(char *filename, DWORD dwOffset, BYTE *buffer, DWORD dwSize);
DWORD CalcCheckSum(BYTE *buffer, DWORD dwSize);

void usage()
{
    printf("Usage: pnewbootloader [-f] filename\n");
}

int main( int argc, char *argv[])
{
    char *filename=NULL;
    bool bForce= false;

    int argsfound=0;
    for (int i=1 ; i<argc ; i++)
    {
        if (argv[i][0]=='-') switch(argv[i][1])
        {
            case 'f': bForce= true; break;
            default:
                usage();
                return 1;
        }
        else switch (argsfound++)
        {
            case 0: filename= argv[i]; break;
        }
    }

    if (argsfound!=1)
    {
        usage();
        return 1;
    }

    BYTE buffer[0x40000]; 
    if (!ReadFromFile(filename, 0, buffer, 0x40000))
    {
        debug("ERROR reading bootloader\n");
        return 1;
    }

    // verify that input file really resembles a bootloader
    if (!bForce && !(buffer[3]==0xea 
            && buffer[64]==69 && buffer[65]==67 && buffer[66]==69 && buffer[67]==67
            && buffer[68]==0 && buffer[69]==0 && buffer[70]==10 && buffer[71]==0x8c))
    {
        debug("this is possibly not a bootloader!!\n");
        hexdump(0x80000000, buffer, 80, 4, 4);
        return 1;
    }

    CheckITSDll();

    DWORD dwFlashInfoOffset;
    if (!FindFlashInfoOffset(dwFlashInfoOffset))
    {
        debug("Unable to find flash info offset, cannot disable bootloader writeprotect\n");
        return 1;
    }
    debug("protection found at %08lx\n", dwFlashInfoOffset);
    if (!DisableWriteProtect(dwFlashInfoOffset))
    {
        debug("unable to disable bootloader writeprotect\n");
        return 1;
    }

    if (!ITFlashROM(0x80000000, buffer, 0x3f000))
        debug("ERROR in ITFlashROM\n");

    if (!EnableWriteProtect(dwFlashInfoOffset))
    {
        debug("unable to re-enable bootloader writeprotect\n");
        return 1;
    }
    CeRapiUninit();

    return 0;
}

bool FindFlashInfoOffset(DWORD& dwFlashInfoOffset)
{
    for (int i=0 ; i<N_VERSIONS ; i++)
    {
        WCHAR vstr[16];
        DWORD nRead;
        if (!ITReadProcessMemory(NULL, versions[i].dwVersion, (BYTE*)vstr, sizeof(vstr), &nRead))
            return false;
        if (memcmp(vstr, versions[i].versionstring, 24))
            continue;

        DWORD addrs[5];
        if (!ITReadProcessMemory(NULL, versions[i].dwAddressList, (BYTE*)addrs, sizeof(addrs), &nRead))
            return false;

        char txt[256];
        if (!ITReadProcessMemory(NULL, addrs[0], (BYTE*)txt, sizeof(txt), &nRead))
            continue;
        if (strcmp(txt, msgFlashStartErase))
            continue;

        dwFlashInfoOffset= versions[i].dwBootloaderAddress;
        return true;

    }
    return false;
}

// this changes the protected area to 60000000 ( and thus opening up 80000000 )
bool DisableWriteProtect(DWORD dwFlashInfoOffset)
{
    DWORD data[3];
    data[0]= 0x60000000;
    data[1]= 0x00040000;
    data[2]= 0x60040000;

    DWORD nWritten;
    return ITWriteProcessMemory(NULL, dwFlashInfoOffset, (BYTE*)data, sizeof(data), &nWritten);
}
// reset it to normal
bool EnableWriteProtect(DWORD dwFlashInfoOffset)
{
    DWORD data[3];
    data[0]= 0x80000000;
    data[1]= 0x00040000;
    data[2]= 0x80040000;

    DWORD nWritten;
    return ITWriteProcessMemory(NULL, dwFlashInfoOffset, (BYTE*)data, sizeof(data), &nWritten);
}
bool ITFlashROM(DWORD dwOffset, BYTE *buffer, DWORD dwBufLength)
{
    DWORD insize= sizeof(FlashROMParams)+dwBufLength;
    FlashROMParams *req= (FlashROMParams*)LocalAlloc(LPTR, insize);

    req->dwOffset= dwOffset;
    req->dwSize= dwBufLength;
    memcpy(req->buffer, buffer, dwBufLength);
    req->dwCheckSum= CalcCheckSum(req->buffer, req->dwSize);

    DWORD outsize=0;
    FlashROMResult *outbuf=NULL;

    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITFlashROM",
            insize, (BYTE*)req,
            &outsize, (BYTE**)&outbuf, NULL, 0);
    if (res)
    {
        error(res, "ITFlashROM(%08lx, %08lx)", dwOffset, dwBufLength);
        return false;
    }
    debug("result: %08lx %08lx\n", outbuf->dwResult, outbuf->dwReturned);
    LocalFree(outbuf);

    return true;
}


bool ReadFromFile(char *filename, DWORD dwOffset, BYTE *buffer, DWORD dwSize)
{
    FILE *f= fopen(filename, "rb");
    if (f==NULL)
    {
        error("fopen(%hs)", filename);
        return false;
    }
    if (0!=fseek(f, dwOffset, SEEK_SET))
    {
        error("fseek(%hs, %08lx)", filename, dwOffset);
        fclose(f);
        return false;
    }
    if (1!=fread(buffer, dwSize, 1, f))
    {
        error("fread(%hs, %08lx)", filename, dwSize);
        fclose(f);
        return false;
    }

    fclose(f);
    return true;
}

DWORD CalcCheckSum(BYTE *buffer, DWORD dwSize)
{
    DWORD sum= 0;
    while (dwSize--)
        sum+=*buffer++;

    return sum;
}

bool ITReadProcessMemory(HANDLE hProc, DWORD dwOffset, BYTE *buffer, DWORD dwBytesWanted, DWORD *pdwNumberOfBytesRead)
{
    ReadProcessMemoryParams inbuf;
    DWORD outsize=0;
    ReadProcessMemoryResult *outbuf=NULL;

    inbuf.hProcess= hProc;
    inbuf.dwOffset= dwOffset;
    inbuf.nSize= dwBytesWanted;
    outbuf= NULL; outsize= 0;
    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITReadProcessMemory",
            sizeof(ReadProcessMemoryParams), (BYTE*)&inbuf,
            &outsize, (BYTE**)&outbuf, NULL, 0);
    if (res)
    {
        error(res, "ITReadProcessMemory");
        return false;
    }
    memcpy(buffer, &outbuf->buffer, outbuf->dwNumberOfBytesRead);
    *pdwNumberOfBytesRead= outbuf->dwNumberOfBytesRead;
    LocalFree(outbuf);

    return true;
}
bool ITWriteProcessMemory(HANDLE hProc, DWORD dwOffset, BYTE *buffer, DWORD dwBufLength, DWORD *pdwNumberOfBytesWritten)
{
    WriteProcessMemoryParams *inbuf;
    DWORD outsize=0;
    WriteProcessMemoryResult *outbuf=NULL;

    DWORD insize= sizeof(WriteProcessMemoryParams)+dwBufLength;
    inbuf= (WriteProcessMemoryParams *)LocalAlloc(LPTR, insize);

    inbuf->hProcess= hProc;
    inbuf->dwOffset= dwOffset;
    inbuf->nSize= dwBufLength;
    memcpy(inbuf->buffer, buffer, dwBufLength);

    outbuf= NULL; outsize= 0;
    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITWriteProcessMemory",
            insize, (BYTE*)inbuf,
            &outsize, (BYTE**)&outbuf, NULL, 0);
    if (res)
    {
        error(res, "ITWriteProcessMemory(%08lx, %08lx, %08lx)", hProc, dwOffset, dwBufLength);
        return false;
    }
    *pdwNumberOfBytesWritten= outbuf->dwNumberOfBytesWritten;
    LocalFree(outbuf);

    return true;
}

